"""Tests for env/base.py - ReasoningEnv abstract base class."""

from datetime import datetime
from typing import Any

import pytest

from chuk_gym_core.env.base import EpisodeState, ReasoningEnv
from chuk_gym_core.schemas.config import (
    DifficultyLevel,
    DifficultyProfile,
    SolverConfig,
    ToolPolicy,
)
from chuk_gym_core.schemas.episode import EpisodeStatus
from chuk_gym_core.schemas.problem import Problem
from chuk_gym_core.schemas.trace import Step, StepOperation, Trace


class ConcreteEnv(ReasoningEnv):
    """Concrete implementation for testing."""

    def __init__(self, **kwargs: Any):
        super().__init__(**kwargs)
        self._grid: list[list[int]] = [[0] * 3 for _ in range(3)]
        self._target: int = 0
        self._current_sum: int = 0

    @property
    def domain(self) -> str:
        return "test_domain"

    @property
    def constraint_types(self) -> list[str]:
        return ["test_constraint"]

    @property
    def business_analogies(self) -> list[str]:
        return ["test_analogy"]

    async def _generate_problem(
        self,
        seed: int,
        difficulty: DifficultyLevel,
    ) -> Problem:
        self._target = seed % 10 + 1
        return Problem(
            id=f"test_{seed}",
            seed=seed,
            domain=self.domain,
            difficulty=difficulty,
            prompt=f"Place numbers to sum to {self._target}",
            gold_answer=str(self._target),
        )

    async def _generate_trace(self, problem: Problem) -> Trace:
        return Trace(
            problem_id=problem.id,
            steps=[
                Step(
                    index=0,
                    operation=StepOperation.PLACE,
                    before_state="empty",
                    after_state=f"sum={self._target}",
                    output_value=self._target,
                )
            ],
        )

    async def _validate_action(self, action: str) -> tuple[bool, float, str]:
        parts = action.split()
        if len(parts) != 2 or parts[0] != "add":
            return False, 0.0, "Invalid action format. Use: add <number>"

        try:
            value = int(parts[1])
        except ValueError:
            return False, 0.0, "Invalid number"

        if value < 0 or value > 10:
            return False, 0.0, "Number must be 0-10"

        self._current_sum += value
        return True, 0.0, f"Added {value}, sum is now {self._current_sum}"

    def _is_complete(self) -> bool:
        return self._current_sum == self._target

    def _get_observation(self) -> dict[str, Any]:
        return {
            "current_sum": self._current_sum,
            "target": self._target,
            "render": f"Sum: {self._current_sum}/{self._target}",
        }

    async def _get_hint(self) -> tuple[Any, str] | None:
        remaining = self._target - self._current_sum
        if remaining > 0:
            return {"value": remaining}, f"Add {remaining} to complete"
        return None

    def _get_legal_actions(self) -> list[str] | None:
        return [f"add {i}" for i in range(11)]

    def _explain_action(self, action: str) -> Step | None:
        return Step(
            index=0,
            operation=StepOperation.PLACE,
            before_state=f"sum={self._current_sum}",
            after_state=action,
            output_value=0,
        )


class TestEpisodeState:
    """Tests for EpisodeState dataclass."""

    def test_create_episode_state(self):
        """Test basic episode state creation."""
        problem = Problem(
            id="test",
            seed=42,
            domain="test",
            difficulty=DifficultyLevel.EASY,
            prompt="Test prompt",
        )
        state = EpisodeState(problem=problem)

        assert state.problem == problem
        assert state.trace is None
        assert state.tool_calls == []
        assert state.moves == []
        assert state.trajectory == []
        assert state.steps_taken == 0
        assert state.invalid_actions == 0
        assert state.hints_used == 0
        assert state.retries == 0
        assert state.done is False
        assert state.final_result is None
        assert isinstance(state.started_at, datetime)
        assert state.last_position is None


class TestReasoningEnv:
    """Tests for ReasoningEnv abstract base class."""

    @pytest.fixture
    def env(self) -> ConcreteEnv:
        return ConcreteEnv()

    @pytest.fixture
    def env_with_config(self) -> ConcreteEnv:
        return ConcreteEnv(
            solver_config=SolverConfig.solver_assisted(budget=5, penalty=0.2),
            max_steps=100,
            step_penalty=-0.01,
            invalid_penalty=-0.5,
            hint_penalty=-0.2,
            correct_reward=1.0,
            completion_bonus=10.0,
            efficiency_multiplier=1.5,
        )

    async def test_reset_basic(self, env: ConcreteEnv):
        """Test basic reset functionality."""
        obs, info = await env.reset(seed=42, difficulty=DifficultyLevel.MEDIUM)

        assert env.state is not None
        assert env.state.problem.seed == 42
        assert env.state.problem.difficulty == DifficultyLevel.MEDIUM
        assert obs["current_sum"] == 0
        assert info["domain"] == "test_domain"

    async def test_reset_with_options(self, env: ConcreteEnv):
        """Test reset with options."""
        custom_config = SolverConfig.solver_free()
        obs, info = await env.reset(
            seed=42,
            difficulty=DifficultyLevel.EASY,
            options={
                "solver_config": custom_config,
                "tool_policy": ToolPolicy.FORBIDDEN,
            },
        )

        assert env.solver_config == custom_config
        assert env.state is not None
        assert env.state.problem.tool_policy == ToolPolicy.FORBIDDEN

    async def test_reset_random_seed(self, env: ConcreteEnv):
        """Test reset with random seed."""
        obs1, _ = await env.reset(seed=None)
        obs2, _ = await env.reset(seed=None)

        # Seeds should be different (very unlikely to be same)
        assert env.state is not None

    async def test_reset_string_difficulty(self, env: ConcreteEnv):
        """Test reset with string difficulty."""
        obs, info = await env.reset(seed=42, difficulty="medium")

        assert env.state is not None
        assert env.state.problem.difficulty == DifficultyLevel.MEDIUM

    async def test_step_valid_action(self, env: ConcreteEnv):
        """Test step with valid action."""
        await env.reset(seed=5, difficulty=DifficultyLevel.EASY)
        # seed=5 gives target=6

        obs, reward, terminated, truncated, info = await env.step("add 3")

        assert info["success"] is True
        assert obs["current_sum"] == 3
        assert terminated is False
        assert truncated is False
        assert env.state is not None
        assert env.state.steps_taken == 1

    async def test_step_invalid_action(self, env: ConcreteEnv):
        """Test step with invalid action."""
        await env.reset(seed=42, difficulty=DifficultyLevel.EASY)

        obs, reward, terminated, truncated, info = await env.step("invalid")

        assert info["success"] is False
        assert env.state is not None
        assert env.state.invalid_actions == 1
        assert reward < 0  # Should include invalid penalty

    async def test_step_completion(self, env: ConcreteEnv):
        """Test step that completes the problem."""
        await env.reset(seed=5, difficulty=DifficultyLevel.EASY)
        # seed=5 gives target=6

        obs, reward, terminated, truncated, info = await env.step("add 6")

        assert terminated is True
        assert env.state is not None
        assert env.state.done is True
        assert reward > 0  # Should include completion bonus

    async def test_step_truncation(self, env: ConcreteEnv):
        """Test step truncation at max steps."""
        env = ConcreteEnv(max_steps=3)
        await env.reset(seed=100, difficulty=DifficultyLevel.EASY)

        for i in range(3):
            obs, reward, terminated, truncated, info = await env.step("add 0")

        assert truncated is True
        assert env.state is not None
        assert env.state.done is True

    async def test_step_without_reset(self, env: ConcreteEnv):
        """Test step without reset raises error."""
        with pytest.raises(RuntimeError, match="Must call reset"):
            await env.step("add 1")

    async def test_step_after_done(self, env: ConcreteEnv):
        """Test step after episode is done raises error."""
        await env.reset(seed=5, difficulty=DifficultyLevel.EASY)
        await env.step("add 6")  # Complete the episode

        with pytest.raises(RuntimeError, match="Episode is done"):
            await env.step("add 1")

    async def test_hint_request(self, env_with_config: ConcreteEnv):
        """Test hint request."""
        await env_with_config.reset(seed=5, difficulty=DifficultyLevel.EASY)

        obs, reward, terminated, truncated, info = await env_with_config.step("hint")

        assert info["success"] is True
        assert "hint" in info
        assert env_with_config.state is not None
        assert env_with_config.state.hints_used == 1
        assert info["hints_remaining"] == 4

    async def test_hint_not_allowed(self, env: ConcreteEnv):
        """Test hint when solver not allowed."""
        env.solver_config = SolverConfig.solver_free()
        await env.reset(seed=42, difficulty=DifficultyLevel.EASY)

        obs, reward, terminated, truncated, info = await env.step("hint")

        assert info["success"] is False
        assert "not allowed" in info["message"]

    async def test_hint_budget_exhausted(self, env: ConcreteEnv):
        """Test hint when budget exhausted."""
        env.solver_config = SolverConfig.solver_assisted(budget=1)
        await env.reset(seed=42, difficulty=DifficultyLevel.EASY)

        # Use the one hint
        await env.step("hint")
        # Try again
        obs, reward, terminated, truncated, info = await env.step("hint")

        assert info["success"] is False
        assert "exhausted" in info["message"]

    async def test_get_episode_record(self, env: ConcreteEnv):
        """Test get_episode_record."""
        await env.reset(seed=99, difficulty=DifficultyLevel.MEDIUM)
        # seed=99 gives target=10 (99 % 10 + 1)
        await env.step("add 1")
        await env.step("add 2")
        # Sum is 3, target is 10, not complete

        record = env.get_episode_record()

        assert record is not None
        assert record.domain == "test_domain"
        assert record.seed == 99
        assert record.steps_taken == 2
        # Episode still in progress - status is ABANDONED since not done
        assert record.status == EpisodeStatus.ABANDONED

    async def test_get_episode_record_solved(self, env: ConcreteEnv):
        """Test get_episode_record for solved episode."""
        await env.reset(seed=5, difficulty=DifficultyLevel.EASY)
        await env.step("add 6")

        record = env.get_episode_record()

        assert record is not None
        assert record.status == EpisodeStatus.SOLVED
        assert record.success is True

    async def test_get_episode_record_timeout(self, env: ConcreteEnv):
        """Test get_episode_record for timeout."""
        env = ConcreteEnv(max_steps=2)
        await env.reset(seed=100, difficulty=DifficultyLevel.EASY)
        await env.step("add 0")
        await env.step("add 0")

        record = env.get_episode_record()

        assert record is not None
        assert record.status == EpisodeStatus.TIMEOUT

    def test_get_episode_record_no_state(self, env: ConcreteEnv):
        """Test get_episode_record with no state."""
        record = env.get_episode_record()
        assert record is None

    async def test_get_canonical_trace(self, env: ConcreteEnv):
        """Test get_canonical_trace."""
        await env.reset(seed=42, difficulty=DifficultyLevel.EASY)

        trace = env.get_canonical_trace()

        assert trace is not None
        assert trace.problem_id.startswith("test_")

    def test_get_canonical_trace_no_state(self, env: ConcreteEnv):
        """Test get_canonical_trace with no state."""
        trace = env.get_canonical_trace()
        assert trace is None

    async def test_current_problem(self, env: ConcreteEnv):
        """Test current_problem property."""
        assert env.current_problem is None

        await env.reset(seed=42, difficulty=DifficultyLevel.EASY)

        assert env.current_problem is not None
        assert env.current_problem.seed == 42

    def test_domain_property(self, env: ConcreteEnv):
        """Test domain property."""
        assert env.domain == "test_domain"

    def test_env_id_property(self, env: ConcreteEnv):
        """Test env_id property."""
        assert env.env_id == "test_domain.v1"

    async def test_render(self, env: ConcreteEnv):
        """Test render method."""
        assert env.render() is None

        await env.reset(seed=42, difficulty=DifficultyLevel.EASY)
        rendered = env.render()

        assert rendered is not None
        assert "Sum:" in rendered

    def test_close(self, env: ConcreteEnv):
        """Test close method."""
        env.close()
        assert env.state is None

    def test_unwrapped(self, env: ConcreteEnv):
        """Test unwrapped property."""
        assert env.unwrapped is env

    def test_available_difficulties(self):
        """Test available_difficulties class method."""
        difficulties = ConcreteEnv.available_difficulties()
        assert len(difficulties) == 7
        assert DifficultyLevel.EASY in difficulties
        assert DifficultyLevel.HARD in difficulties

    async def test_get_difficulty_profile(self, env: ConcreteEnv):
        """Test get_difficulty_profile."""
        profile = env.get_difficulty_profile()
        assert isinstance(profile, DifficultyProfile)

        await env.reset(seed=42, difficulty=DifficultyLevel.HARD)
        profile = env.get_difficulty_profile()
        assert profile.logic_depth >= 1

    async def test_constraint_types(self, env: ConcreteEnv):
        """Test constraint_types property."""
        assert env.constraint_types == ["test_constraint"]

    async def test_business_analogies(self, env: ConcreteEnv):
        """Test business_analogies property."""
        assert env.business_analogies == ["test_analogy"]

    async def test_optimal_steps(self, env: ConcreteEnv):
        """Test optimal_steps property."""
        assert env.optimal_steps is None

        await env.reset(seed=42, difficulty=DifficultyLevel.EASY)
        assert env.optimal_steps == 1  # Our trace has 1 step

    async def test_get_legal_actions(self, env: ConcreteEnv):
        """Test _get_legal_actions."""
        actions = env._get_legal_actions()
        assert actions is not None
        assert len(actions) == 11

    async def test_explain_action(self, env: ConcreteEnv):
        """Test _explain_action."""
        await env.reset(seed=42, difficulty=DifficultyLevel.EASY)
        step = env._explain_action("add 5")

        assert step is not None
        assert step.operation == StepOperation.PLACE

    async def test_move_recording(self, env: ConcreteEnv):
        """Test that moves are recorded."""
        await env.reset(seed=42, difficulty=DifficultyLevel.EASY)
        await env.step("add 1")
        await env.step("add 2")

        assert env.state is not None
        assert len(env.state.moves) == 2
        assert env.state.moves[0].action == "add 1"
        assert env.state.moves[1].action == "add 2"

    async def test_efficiency_bonus(self, env: ConcreteEnv):
        """Test efficiency bonus calculation."""
        env = ConcreteEnv(
            completion_bonus=10.0,
            efficiency_multiplier=2.0,
        )
        await env.reset(seed=5, difficulty=DifficultyLevel.EASY)

        # Complete in optimal steps (1 step, trace has 1 step)
        obs, reward, terminated, truncated, info = await env.step("add 6")

        # Should get full efficiency bonus
        assert reward > 10.0  # completion_bonus * efficiency * multiplier


class TestEnvWithoutHints:
    """Tests for env without hint implementation."""

    class NoHintEnv(ConcreteEnv):
        async def _get_hint(self) -> tuple[Any, str] | None:
            return None

    async def test_hint_not_available(self):
        """Test when hint implementation returns None."""
        env = self.NoHintEnv()
        await env.reset(seed=42, difficulty=DifficultyLevel.EASY)

        obs, reward, terminated, truncated, info = await env.step("hint")

        assert info["success"] is False
        assert "No hint available" in info["message"]



"""Tests for export/jsonl.py and export/formats.py."""

import json
from datetime import datetime
from io import StringIO
from pathlib import Path
from tempfile import NamedTemporaryFile

import pytest

from chuk_gym_core.export.formats import ExportFormat
from chuk_gym_core.export.jsonl import JSONLExporter
from chuk_gym_core.schemas.config import DifficultyLevel
from chuk_gym_core.schemas.episode import (
    EpisodeRecord,
    EpisodeStatus,
    MoveRecord,
    TrajectoryStep,
)
from chuk_gym_core.schemas.problem import Problem
from chuk_gym_core.schemas.trace import Step, StepOperation, Trace


class TestExportFormat:
    """Tests for ExportFormat enum."""

    def test_all_formats_exist(self):
        """Test all expected formats exist."""
        assert ExportFormat.JSONL == "jsonl"
        assert ExportFormat.JSON == "json"
        assert ExportFormat.CSV == "csv"
        assert ExportFormat.PARQUET == "parquet"
        assert ExportFormat.CHAT == "chat"
        assert ExportFormat.QA == "qa"
        assert ExportFormat.INSTRUCT == "instruct"
        assert ExportFormat.HF_DATASET == "hf_dataset"

    def test_format_values(self):
        """Test format string values."""
        assert ExportFormat.JSONL.value == "jsonl"
        assert ExportFormat.CHAT.value == "chat"


class TestJSONLExporter:
    """Tests for JSONLExporter."""

    @pytest.fixture
    def problem(self) -> Problem:
        return Problem(
            id="test_1",
            seed=42,
            domain="arithmetic",
            difficulty=DifficultyLevel.MEDIUM,
            prompt="What is 2 + 2?",
            expression="2 + 2",
            gold_answer="4",
            initial_state={"expression": "2 + 2"},
            constraint_types=["addition"],
        )

    @pytest.fixture
    def trace(self) -> Trace:
        return Trace(
            problem_id="test_1",
            steps=[
                Step(
                    index=0,
                    operation=StepOperation.ADD,
                    before_state="2 + 2",
                    after_state="4",
                    output_value=4,
                    explanation="Add 2 and 2 to get 4",
                )
            ],
        )

    @pytest.fixture
    def episode(self) -> EpisodeRecord:
        now = datetime.now()
        return EpisodeRecord(
            episode_id="ep_test_1",
            env_id="arithmetic.v1",
            instance_id="seed:42/diff:medium",
            domain="arithmetic",
            difficulty=DifficultyLevel.MEDIUM,
            seed=42,
            prompt="What is 2 + 2?",
            started_at=now,
            ended_at=now,
            wall_time_ms=1000,
            status=EpisodeStatus.SOLVED,
            final_answer="4",
            gold_answer="4",
            steps_taken=1,
            invalid_actions=0,
            hints_used=0,
            optimal_steps=1,
            trajectory=[
                TrajectoryStep(
                    t=0,
                    observation={"state": "2 + 2"},
                    action="ANSWER 4",
                    reward=1.0,
                    next_observation={"state": "solved"},
                    done=True,
                    teacher_steps=[
                        Step(
                            index=0,
                            operation=StepOperation.ADD,
                            before_state="2 + 2",
                            after_state="4",
                            output_value=4,
                        )
                    ],
                )
            ],
            move_history=[
                MoveRecord(
                    step=0,
                    action="ANSWER 4",
                    success=True,
                    reward=1.0,
                )
            ],
        )

    def test_init_with_string_path(self):
        """Test initialization with string path."""
        with NamedTemporaryFile(mode="w", suffix=".jsonl", delete=False) as f:
            path = f.name

        exporter = JSONLExporter(path)
        exporter.close()

        Path(path).unlink()

    def test_init_with_path_object(self):
        """Test initialization with Path object."""
        with NamedTemporaryFile(mode="w", suffix=".jsonl", delete=False) as f:
            path = Path(f.name)

        exporter = JSONLExporter(path)
        exporter.close()

        path.unlink()

    def test_init_with_file_handle(self):
        """Test initialization with file handle."""
        output = StringIO()
        exporter = JSONLExporter(output)
        exporter.close()

    def test_context_manager(self, problem: Problem, trace: Trace):
        """Test context manager usage."""
        output = StringIO()

        with JSONLExporter(output) as exporter:
            exporter.write_problem(problem, trace)

        assert exporter.count == 1

    def test_write_problem(self, problem: Problem, trace: Trace):
        """Test write_problem."""
        output = StringIO()
        exporter = JSONLExporter(output)
        exporter.write_problem(problem, trace)
        exporter.close()

        output.seek(0)
        line = output.readline()
        data = json.loads(line)

        assert data["type"] == "problem"
        assert data["id"] == "test_1"
        assert data["domain"] == "arithmetic"
        assert data["difficulty"] == "medium"
        assert data["seed"] == 42
        assert data["prompt"] == "What is 2 + 2?"
        assert data["expression"] == "2 + 2"
        assert data["gold_answer"] == "4"
        assert data["initial_state"] == {"expression": "2 + 2"}
        assert data["constraint_types"] == ["addition"]
        assert "trace" in data
        assert data["optimal_steps"] == 1

    def test_write_problem_without_trace(self, problem: Problem):
        """Test write_problem without trace."""
        output = StringIO()
        exporter = JSONLExporter(output)
        exporter.write_problem(problem)
        exporter.close()

        output.seek(0)
        data = json.loads(output.readline())

        assert "trace" not in data
        assert "optimal_steps" not in data

    def test_write_problem_include_trace_false(self, problem: Problem, trace: Trace):
        """Test write_problem with include_trace=False."""
        output = StringIO()
        exporter = JSONLExporter(output, include_trace=False)
        exporter.write_problem(problem, trace)
        exporter.close()

        output.seek(0)
        data = json.loads(output.readline())

        assert "trace" not in data

    def test_write_problem_minimal(self):
        """Test write_problem with minimal problem."""
        problem = Problem(
            id="minimal",
            seed=1,
            domain="test",
            difficulty=DifficultyLevel.EASY,
            prompt="Minimal test",
        )

        output = StringIO()
        exporter = JSONLExporter(output)
        exporter.write_problem(problem)
        exporter.close()

        output.seek(0)
        data = json.loads(output.readline())

        assert data["id"] == "minimal"
        assert "expression" not in data
        assert "gold_answer" not in data

    def test_write_episode(self, episode: EpisodeRecord):
        """Test write_episode."""
        output = StringIO()
        exporter = JSONLExporter(output)
        exporter.write_episode(episode)
        exporter.close()

        output.seek(0)
        data = json.loads(output.readline())

        assert data["type"] == "episode"
        assert data["episode_id"] == "ep_test_1"
        assert data["env_id"] == "arithmetic.v1"
        assert data["domain"] == "arithmetic"
        assert data["difficulty"] == "medium"
        assert data["status"] == "solved"
        assert data["success"] is True
        assert data["steps_taken"] == 1
        assert data["efficiency_score"] == 1.0
        assert "trajectory" in data

    def test_write_episode_without_trajectory(self):
        """Test write_episode without trajectory."""
        now = datetime.now()
        episode = EpisodeRecord(
            episode_id="ep_test",
            env_id="test.v1",
            instance_id="test",
            domain="test",
            difficulty=DifficultyLevel.EASY,
            seed=1,
            prompt="Test",
            started_at=now,
            ended_at=now,
            wall_time_ms=500,
            status=EpisodeStatus.FAILED,
            steps_taken=5,
            invalid_actions=2,
            hints_used=0,
        )

        output = StringIO()
        exporter = JSONLExporter(output)
        exporter.write_episode(episode)
        exporter.close()

        output.seek(0)
        data = json.loads(output.readline())

        assert "trajectory" not in data
        assert data["success"] is False

    def test_write_episode_include_teacher_steps_false(self, episode: EpisodeRecord):
        """Test write_episode with include_teacher_steps=False."""
        output = StringIO()
        exporter = JSONLExporter(output, include_teacher_steps=False)
        exporter.write_episode(episode)
        exporter.close()

        output.seek(0)
        data = json.loads(output.readline())

        assert "trajectory" not in data

    def test_write_training_example_qa(self, problem: Problem, trace: Trace):
        """Test write_training_example with QA format."""
        output = StringIO()
        exporter = JSONLExporter(output)
        exporter.write_training_example(problem, trace, format_type="qa")
        exporter.close()

        output.seek(0)
        data = json.loads(output.readline())

        assert data["type"] == "training_qa"
        assert data["question"] == "What is 2 + 2?"
        assert data["answer"] == "4"
        assert "steps" in data
        assert data["metadata"]["id"] == "test_1"
        assert data["metadata"]["domain"] == "arithmetic"
        assert data["metadata"]["difficulty"] == "medium"
        assert data["metadata"]["seed"] == 42

    def test_write_training_example_chat(self, problem: Problem, trace: Trace):
        """Test write_training_example with chat format."""
        output = StringIO()
        exporter = JSONLExporter(output)
        exporter.write_training_example(problem, trace, format_type="chat")
        exporter.close()

        output.seek(0)
        data = json.loads(output.readline())

        assert data["type"] == "training_chat"
        assert "messages" in data
        assert len(data["messages"]) == 2
        assert data["messages"][0]["role"] == "user"
        assert data["messages"][1]["role"] == "assistant"
        assert "Final answer: 4" in data["messages"][1]["content"]

    def test_write_training_example_instruct(self, problem: Problem, trace: Trace):
        """Test write_training_example with instruct format."""
        output = StringIO()
        exporter = JSONLExporter(output)
        exporter.write_training_example(problem, trace, format_type="instruct")
        exporter.close()

        output.seek(0)
        data = json.loads(output.readline())

        assert data["type"] == "training_instruct"
        assert "instruction" in data
        assert "arithmetic" in data["instruction"]
        assert data["input"] == "What is 2 + 2?"
        assert "Answer: 4" in data["output"]

    def test_write_training_example_invalid_format(self, problem: Problem, trace: Trace):
        """Test write_training_example with invalid format."""
        output = StringIO()
        exporter = JSONLExporter(output)

        with pytest.raises(ValueError, match="Unknown format type"):
            exporter.write_training_example(problem, trace, format_type="invalid")

    def test_count_property(self, problem: Problem, trace: Trace):
        """Test count property."""
        output = StringIO()
        exporter = JSONLExporter(output)

        assert exporter.count == 0

        exporter.write_problem(problem, trace)
        assert exporter.count == 1

        exporter.write_problem(problem, trace)
        assert exporter.count == 2

        exporter.close()

    def test_flush(self, problem: Problem, trace: Trace):
        """Test flush method."""
        output = StringIO()
        exporter = JSONLExporter(output)
        exporter.write_problem(problem, trace)
        exporter.flush()

        output.seek(0)
        assert len(output.readline()) > 0

        exporter.close()

    def test_multiple_records(self, problem: Problem, trace: Trace):
        """Test writing multiple records."""
        output = StringIO()
        exporter = JSONLExporter(output)

        for i in range(5):
            p = Problem(
                id=f"test_{i}",
                seed=i,
                domain="test",
                difficulty=DifficultyLevel.EASY,
                prompt=f"Problem {i}",
            )
            exporter.write_problem(p)

        exporter.close()

        output.seek(0)
        lines = output.readlines()
        assert len(lines) == 5

        for i, line in enumerate(lines):
            data = json.loads(line)
            assert data["id"] == f"test_{i}"

    def test_write_problem_with_none_values(self):
        """Test write_problem handles None values correctly."""
        problem = Problem(
            id="test",
            seed=1,
            domain="test",
            difficulty=DifficultyLevel.EASY,
            prompt="Test",
            expression=None,
            gold_answer=None,
            initial_state=None,
        )

        output = StringIO()
        exporter = JSONLExporter(output)
        exporter.write_problem(problem)
        exporter.close()

        output.seek(0)
        data = json.loads(output.readline())

        assert "expression" not in data
        assert "gold_answer" not in data
        assert "initial_state" not in data


class TestJSONLExporterFileOperations:
    """Tests for JSONLExporter file operations."""

    def test_write_to_file(self):
        """Test writing to actual file."""
        with NamedTemporaryFile(mode="w", suffix=".jsonl", delete=False) as f:
            path = f.name

        problem = Problem(
            id="file_test",
            seed=1,
            domain="test",
            difficulty=DifficultyLevel.EASY,
            prompt="File test",
        )

        with JSONLExporter(path) as exporter:
            exporter.write_problem(problem)

        # Read and verify
        with open(path, "r") as f:
            data = json.loads(f.readline())

        assert data["id"] == "file_test"

        Path(path).unlink()

    def test_close_idempotent(self):
        """Test that close can be called multiple times."""
        output = StringIO()
        exporter = JSONLExporter(output)
        exporter.close()
        exporter.close()  # Should not raise

    def test_write_after_close_file(self):
        """Test that write after close to file does nothing."""
        with NamedTemporaryFile(mode="w", suffix=".jsonl", delete=False) as f:
            path = f.name

        exporter = JSONLExporter(path)
        exporter.close()

        # After close, _output is None, so write does nothing
        problem = Problem(
            id="test",
            seed=1,
            domain="test",
            difficulty=DifficultyLevel.EASY,
            prompt="Test",
        )
        exporter.write_problem(problem)

        # File should be empty
        with open(path, "r") as f:
            assert f.read() == ""

        Path(path).unlink()



"""
Export format definitions for dataset generation.
"""

from enum import Enum


class ExportFormat(str, Enum):
    """Supported export formats for datasets."""

    JSONL = "jsonl"  # JSON Lines (one record per line)
    JSON = "json"  # Pretty-printed JSON array
    CSV = "csv"  # Comma-separated values
    PARQUET = "parquet"  # Apache Parquet (columnar)

    # Training-specific formats
    CHAT = "chat"  # Chat/conversation format
    QA = "qa"  # Question-answer pairs
    INSTRUCT = "instruct"  # Instruction-following format

    # Framework-specific
    HF_DATASET = "hf_dataset"  # Hugging Face datasets format



"""Tests for curriculum scheduler."""

import pytest

from chuk_gym_core.curriculum.scheduler import (
    CurriculumScheduler,
    LinearProgression,
    PerformanceBasedProgression,
    PerformanceMetrics,
    StepBasedProgression,
)
from chuk_gym_core.schemas.config import DifficultyLevel


class TestPerformanceMetrics:
    """Tests for PerformanceMetrics."""

    def test_initial_state(self):
        """Test initial metrics state."""
        metrics = PerformanceMetrics()
        assert metrics.total_episodes == 0
        assert metrics.solve_rate == 0.0
        assert metrics.error_rate == 0.0

    def test_update(self):
        """Test metrics update."""
        metrics = PerformanceMetrics()
        metrics.update(solved=True, steps=10, invalid=2, hints=1, efficiency=0.8)

        assert metrics.total_episodes == 1
        assert metrics.solved_episodes == 1
        assert metrics.total_steps == 10
        assert metrics.total_invalid == 2

    def test_solve_rate(self):
        """Test solve rate calculation."""
        metrics = PerformanceMetrics()
        metrics.update(solved=True, steps=10, invalid=0, hints=0, efficiency=1.0)
        metrics.update(solved=False, steps=5, invalid=0, hints=0, efficiency=0.0)

        assert metrics.solve_rate == 0.5

    def test_error_rate_calculation(self):
        """Test error rate calculation."""
        metrics = PerformanceMetrics(total_steps=80, total_invalid=20)
        assert metrics.error_rate == 0.2

    def test_error_rate_with_zero_actions(self):
        """Test error rate returns 0 when no actions."""
        metrics = PerformanceMetrics()
        assert metrics.error_rate == 0.0

    def test_efficiency_first_episode(self):
        """Test efficiency is set directly on first episode."""
        metrics = PerformanceMetrics()
        metrics.update(solved=True, steps=10, invalid=0, hints=0, efficiency=0.8)
        assert metrics.avg_efficiency == 0.8

    def test_efficiency_exponential_average(self):
        """Test efficiency uses exponential moving average."""
        metrics = PerformanceMetrics()
        metrics.update(solved=True, steps=10, invalid=0, hints=0, efficiency=1.0)
        metrics.update(solved=True, steps=10, invalid=0, hints=0, efficiency=0.5)
        # EMA: 0.1 * 0.5 + 0.9 * 1.0 = 0.95
        assert metrics.avg_efficiency == pytest.approx(0.95)


class TestLinearProgression:
    """Tests for LinearProgression strategy."""

    def test_progression(self):
        """Test linear difficulty progression."""
        strategy = LinearProgression(
            episodes_per_level=10,
            start_level=DifficultyLevel.EASY,
        )
        metrics = PerformanceMetrics()

        # Should start at easy
        level = strategy.get_difficulty(DifficultyLevel.EASY, metrics, step=0)
        assert level == DifficultyLevel.EASY

        # After 10 episodes, should advance
        level = strategy.get_difficulty(DifficultyLevel.EASY, metrics, step=10)
        assert level == DifficultyLevel.PRETTY_EASY

        # After 20 episodes, should advance again
        level = strategy.get_difficulty(DifficultyLevel.EASY, metrics, step=20)
        assert level == DifficultyLevel.MEDIUM

    def test_should_advance_on_schedule(self):
        """Test should_advance returns True when on schedule."""
        strategy = LinearProgression(episodes_per_level=10)
        metrics = PerformanceMetrics(total_episodes=10)
        assert strategy.should_advance(DifficultyLevel.EASY, metrics) is True

    def test_should_advance_not_on_schedule(self):
        """Test should_advance returns False when not on schedule."""
        strategy = LinearProgression(episodes_per_level=10)
        metrics = PerformanceMetrics(total_episodes=7)
        assert strategy.should_advance(DifficultyLevel.EASY, metrics) is False

    def test_max_level_cap(self):
        """Test difficulty is capped at maximum level."""
        strategy = LinearProgression(episodes_per_level=10)
        metrics = PerformanceMetrics()
        # Very high step should still cap at max level
        level = strategy.get_difficulty(DifficultyLevel.EASY, metrics, step=1000)
        assert level == DifficultyLevel.VERY_HARD


class TestPerformanceBasedProgression:
    """Tests for PerformanceBasedProgression strategy."""

    def test_advance_on_high_performance(self):
        """Test advancement when solve rate is high."""
        strategy = PerformanceBasedProgression(
            advance_threshold=0.8,
            min_episodes=5,
        )

        metrics = PerformanceMetrics()
        # Simulate high success rate
        for _ in range(10):
            metrics.update(solved=True, steps=5, invalid=0, hints=0, efficiency=1.0)

        assert metrics.solve_rate >= 0.8
        assert strategy.should_advance(DifficultyLevel.EASY, metrics)

    def test_no_advance_on_low_performance(self):
        """Test no advancement when solve rate is low."""
        strategy = PerformanceBasedProgression(
            advance_threshold=0.8,
            min_episodes=5,
        )

        metrics = PerformanceMetrics()
        # Simulate low success rate
        for _ in range(10):
            metrics.update(solved=False, steps=5, invalid=2, hints=0, efficiency=0.0)

        assert metrics.solve_rate < 0.8
        assert not strategy.should_advance(DifficultyLevel.EASY, metrics)

    def test_retreat_on_very_low_performance(self):
        """Test retreat when solve rate drops."""
        strategy = PerformanceBasedProgression(
            retreat_threshold=0.3,
            min_episodes=5,
        )

        metrics = PerformanceMetrics()
        # Simulate very low success rate
        for _ in range(10):
            metrics.update(solved=False, steps=5, invalid=2, hints=0, efficiency=0.0)

        assert metrics.solve_rate <= 0.3
        assert strategy.should_retreat(DifficultyLevel.MEDIUM, metrics)

    def test_get_difficulty_below_min_episodes(self):
        """Test get_difficulty returns current when below min episodes."""
        strategy = PerformanceBasedProgression(min_episodes=20)
        metrics = PerformanceMetrics(total_episodes=10, solved_episodes=10)

        result = strategy.get_difficulty(DifficultyLevel.EASY, metrics, step=10)
        assert result == DifficultyLevel.EASY

    def test_get_difficulty_advance(self):
        """Test get_difficulty advances on high solve rate."""
        strategy = PerformanceBasedProgression(advance_threshold=0.8, min_episodes=10)
        metrics = PerformanceMetrics(total_episodes=15, solved_episodes=14)

        result = strategy.get_difficulty(DifficultyLevel.EASY, metrics, step=15)
        assert result == DifficultyLevel.PRETTY_EASY

    def test_get_difficulty_retreat(self):
        """Test get_difficulty retreats on low solve rate."""
        strategy = PerformanceBasedProgression(retreat_threshold=0.3, min_episodes=10)
        metrics = PerformanceMetrics(total_episodes=15, solved_episodes=2)

        result = strategy.get_difficulty(DifficultyLevel.MEDIUM, metrics, step=15)
        assert result == DifficultyLevel.PRETTY_EASY

    def test_get_difficulty_no_change(self):
        """Test get_difficulty stays same when mid-range solve rate."""
        strategy = PerformanceBasedProgression(
            advance_threshold=0.8, retreat_threshold=0.3, min_episodes=10
        )
        metrics = PerformanceMetrics(total_episodes=15, solved_episodes=8)

        result = strategy.get_difficulty(DifficultyLevel.MEDIUM, metrics, step=15)
        assert result == DifficultyLevel.MEDIUM

    def test_no_retreat_from_lowest(self):
        """Test no retreat from lowest difficulty."""
        strategy = PerformanceBasedProgression(retreat_threshold=0.3, min_episodes=10)
        metrics = PerformanceMetrics(total_episodes=15, solved_episodes=0)

        result = strategy.get_difficulty(DifficultyLevel.VERY_EASY, metrics, step=15)
        assert result == DifficultyLevel.VERY_EASY

    def test_no_advance_from_highest(self):
        """Test no advance from highest difficulty."""
        strategy = PerformanceBasedProgression(advance_threshold=0.8, min_episodes=10)
        metrics = PerformanceMetrics(total_episodes=15, solved_episodes=15)

        result = strategy.get_difficulty(DifficultyLevel.VERY_HARD, metrics, step=15)
        assert result == DifficultyLevel.VERY_HARD

    def test_should_advance_below_min(self):
        """Test should_advance returns False below min episodes."""
        strategy = PerformanceBasedProgression(min_episodes=20)
        metrics = PerformanceMetrics(total_episodes=10, solved_episodes=10)
        assert strategy.should_advance(DifficultyLevel.EASY, metrics) is False

    def test_should_retreat_below_min(self):
        """Test should_retreat returns False below min episodes."""
        strategy = PerformanceBasedProgression(min_episodes=20)
        metrics = PerformanceMetrics(total_episodes=10, solved_episodes=0)
        assert strategy.should_retreat(DifficultyLevel.MEDIUM, metrics) is False

    def test_record_result(self):
        """Test record_result adds to window."""
        strategy = PerformanceBasedProgression(window_size=5)
        strategy.record_result(True)
        strategy.record_result(True)
        strategy.record_result(False)

        assert len(strategy._window_results) == 3
        assert strategy.windowed_solve_rate == pytest.approx(2 / 3)

    def test_windowed_solve_rate_empty(self):
        """Test windowed_solve_rate returns 0 for empty window."""
        strategy = PerformanceBasedProgression()
        assert strategy.windowed_solve_rate == 0.0

    def test_window_size_limit(self):
        """Test window is limited to window_size."""
        strategy = PerformanceBasedProgression(window_size=3)
        for _ in range(5):
            strategy.record_result(True)

        assert len(strategy._window_results) == 3


class TestStepBasedProgression:
    """Tests for StepBasedProgression strategy."""

    def test_advance_after_required_solves(self):
        """Test advancement after consecutive solves."""
        strategy = StepBasedProgression(required_solves=3)

        # Record 3 consecutive solves
        for _ in range(3):
            strategy.record_result(solved=True)

        metrics = PerformanceMetrics()
        assert strategy.should_advance(DifficultyLevel.EASY, metrics)

    def test_reset_on_failure(self):
        """Test streak reset on failure."""
        strategy = StepBasedProgression(required_solves=3)

        # Record 2 solves then fail
        strategy.record_result(solved=True)
        strategy.record_result(solved=True)
        strategy.record_result(solved=False)

        metrics = PerformanceMetrics()
        assert not strategy.should_advance(DifficultyLevel.EASY, metrics)

    def test_get_difficulty_advances(self):
        """Test get_difficulty advances when should_advance is True."""
        strategy = StepBasedProgression(required_solves=2)
        for _ in range(2):
            strategy.record_result(solved=True)
        metrics = PerformanceMetrics()

        result = strategy.get_difficulty(DifficultyLevel.EASY, metrics, step=2)
        assert result == DifficultyLevel.PRETTY_EASY
        # Check that internal counters are reset
        assert strategy._consecutive_solves == 0
        assert strategy._attempts_at_level == 0

    def test_get_difficulty_no_advance(self):
        """Test get_difficulty returns current when not advancing."""
        strategy = StepBasedProgression(required_solves=10)
        strategy.record_result(solved=True)
        metrics = PerformanceMetrics()

        result = strategy.get_difficulty(DifficultyLevel.EASY, metrics, step=1)
        assert result == DifficultyLevel.EASY

    def test_forced_advance_on_max_attempts(self):
        """Test forced advancement after max attempts."""
        strategy = StepBasedProgression(max_attempts_per_level=5)
        for _ in range(5):
            strategy.record_result(solved=False)
        metrics = PerformanceMetrics()

        assert strategy.should_advance(DifficultyLevel.EASY, metrics) is True

    def test_no_advance_past_highest(self):
        """Test no advance past highest difficulty."""
        strategy = StepBasedProgression(required_solves=1)
        strategy.record_result(solved=True)
        metrics = PerformanceMetrics()

        result = strategy.get_difficulty(DifficultyLevel.VERY_HARD, metrics, step=1)
        assert result == DifficultyLevel.VERY_HARD

    def test_reset_level(self):
        """Test reset_level resets counters."""
        strategy = StepBasedProgression()
        strategy.record_result(solved=True)
        strategy.record_result(solved=True)

        strategy.reset_level()

        assert strategy._consecutive_solves == 0
        assert strategy._attempts_at_level == 0


class TestCurriculumScheduler:
    """Tests for CurriculumScheduler."""

    def test_basic_usage(self):
        """Test basic scheduler usage."""
        scheduler = CurriculumScheduler(
            strategy=LinearProgression(episodes_per_level=5, start_level=DifficultyLevel.EASY),
            start_difficulty=DifficultyLevel.EASY,
        )

        assert scheduler.get_current_difficulty() == DifficultyLevel.EASY

        # Record some episodes
        for _ in range(5):
            scheduler.record_episode(solved=True, steps=10)

        # Should have advanced (LinearProgression advances on step count, not episode count)
        # After 5 episodes at step 5, get_difficulty with step=5 returns PRETTY_EASY
        assert scheduler.get_current_difficulty() == DifficultyLevel.PRETTY_EASY

    def test_get_summary(self):
        """Test summary generation."""
        scheduler = CurriculumScheduler(
            strategy=LinearProgression(episodes_per_level=10, start_level=DifficultyLevel.EASY),
            start_difficulty=DifficultyLevel.EASY,
        )

        for i in range(5):
            scheduler.record_episode(
                solved=i % 2 == 0,
                steps=10,
                invalid=1,
                hints=0,
                efficiency=0.8,
            )

        summary = scheduler.get_summary()
        assert summary["total_episodes"] == 5
        assert summary["current_difficulty"] == DifficultyLevel.EASY.value
        assert 0 <= summary["solve_rate"] <= 1

    def test_reset(self):
        """Test scheduler reset."""
        scheduler = CurriculumScheduler(
            strategy=LinearProgression(episodes_per_level=5),
            start_difficulty=DifficultyLevel.MEDIUM,
        )

        for _ in range(10):
            scheduler.record_episode(solved=True, steps=5)

        scheduler.reset()

        assert scheduler.get_current_difficulty() == DifficultyLevel.MEDIUM
        assert scheduler.metrics.total_episodes == 0
        assert scheduler.step == 0

    def test_record_episode_calls_strategy_record_result(self):
        """Test record_episode calls strategy.record_result if available."""
        strategy = StepBasedProgression(required_solves=3)
        scheduler = CurriculumScheduler(strategy=strategy)

        scheduler.record_episode(solved=True)
        assert strategy._consecutive_solves == 1

        scheduler.record_episode(solved=True)
        assert strategy._consecutive_solves == 2

    def test_record_episode_triggers_reset_level_on_advance(self):
        """Test reset_level is called when difficulty advances."""
        strategy = StepBasedProgression(required_solves=2)
        scheduler = CurriculumScheduler(strategy=strategy)

        scheduler.record_episode(solved=True)
        scheduler.record_episode(solved=True)

        # After advancing, reset_level should have been called
        assert strategy._consecutive_solves == 0
        assert strategy._attempts_at_level == 0

    def test_reset_calls_strategy_reset_level(self):
        """Test reset calls strategy.reset_level if available."""
        strategy = StepBasedProgression()
        scheduler = CurriculumScheduler(strategy=strategy)

        scheduler.record_episode(solved=True)
        scheduler.reset()

        assert strategy._consecutive_solves == 0
        assert strategy._attempts_at_level == 0

    def test_record_episode_updates_history(self):
        """Test record_episode updates history."""
        scheduler = CurriculumScheduler(
            strategy=LinearProgression(episodes_per_level=100),
        )

        scheduler.record_episode(solved=True)

        assert len(scheduler.history) == 1
        assert scheduler.history[0]["step"] == 1
        assert scheduler.history[0]["solved"] is True

    def test_get_summary_all_fields(self):
        """Test get_summary returns all expected fields."""
        scheduler = CurriculumScheduler(strategy=LinearProgression())
        scheduler.record_episode(solved=True, steps=10, invalid=2, hints=1, efficiency=0.8)

        summary = scheduler.get_summary()

        assert "current_difficulty" in summary
        assert "total_episodes" in summary
        assert "solve_rate" in summary
        assert "error_rate" in summary
        assert "avg_efficiency" in summary

    def test_strategy_without_record_result(self):
        """Test scheduler works with strategy without record_result."""
        scheduler = CurriculumScheduler(
            strategy=LinearProgression(episodes_per_level=5),
        )

        # Should not raise
        scheduler.record_episode(solved=True)
        assert scheduler.metrics.total_episodes == 1

    def test_strategy_without_reset_level(self):
        """Test reset works with strategy without reset_level."""
        scheduler = CurriculumScheduler(
            strategy=LinearProgression(episodes_per_level=5),
        )

        for _ in range(5):
            scheduler.record_episode(solved=True)

        # Should not raise
        scheduler.reset()
        assert scheduler.step == 0



"""Tests for verifiers/base.py - Verifier abstract base class."""

from typing import Any

import pytest

from chuk_gym_core.schemas.config import DifficultyLevel
from chuk_gym_core.schemas.problem import Problem
from chuk_gym_core.schemas.trace import Step, StepOperation, StepRef, Trace
from chuk_gym_core.schemas.verification import (
    ErrorType,
    VerificationResult,
)
from chuk_gym_core.verifiers.base import Verifier


class ConcreteVerifier(Verifier):
    """Concrete implementation for testing."""

    @property
    def domain(self) -> str:
        return "test_verifier"

    def verify_final(
        self,
        problem: Problem,
        candidate: str,
    ) -> VerificationResult:
        if problem.gold_answer is None:
            return VerificationResult.failure(
                ErrorType.INCOMPLETE,
                "No gold answer available",
            )

        if candidate.strip() == problem.gold_answer.strip():
            return VerificationResult.success(
                expected=problem.gold_answer,
                actual=candidate,
            )

        return VerificationResult.failure(
            ErrorType.WRONG_ANSWER,
            f"Expected {problem.gold_answer}, got {candidate}",
            expected=problem.gold_answer,
            actual=candidate,
        )


class NumericVerifier(Verifier):
    """Verifier for numeric answers with tolerance."""

    @property
    def domain(self) -> str:
        return "numeric_verifier"

    def verify_final(
        self,
        problem: Problem,
        candidate: str,
    ) -> VerificationResult:
        try:
            candidate_value = float(candidate)
        except ValueError:
            return VerificationResult.failure(
                ErrorType.INVALID_FORMAT,
                "Could not parse as number",
            )

        if problem.gold_answer is None:
            return VerificationResult.failure(ErrorType.INCOMPLETE)

        gold_value = float(problem.gold_answer)
        tolerance = problem.answer_tolerance or 1e-9

        if abs(candidate_value - gold_value) < tolerance:
            return VerificationResult.success(expected=gold_value, actual=candidate_value)

        return VerificationResult.failure(
            ErrorType.WRONG_ANSWER,
            expected=gold_value,
            actual=candidate_value,
        )

    def classify_error(
        self,
        problem: Problem,
        expected: Any,
        actual: Any,
    ) -> ErrorType:
        if actual is None:
            return ErrorType.INCOMPLETE

        if isinstance(expected, (int, float)) and isinstance(actual, (int, float)):
            if abs(expected) > 0 and abs(expected - actual) / abs(expected) < 0.01:
                return ErrorType.ROUNDING_ERROR
            if expected == -actual:
                return ErrorType.SIGN_ERROR

        return ErrorType.WRONG_ANSWER


class TestVerifier:
    """Tests for Verifier abstract base class."""

    @pytest.fixture
    def verifier(self) -> ConcreteVerifier:
        return ConcreteVerifier()

    @pytest.fixture
    def numeric_verifier(self) -> NumericVerifier:
        return NumericVerifier()

    @pytest.fixture
    def problem(self) -> Problem:
        return Problem(
            id="test_1",
            seed=42,
            domain="test",
            difficulty=DifficultyLevel.MEDIUM,
            prompt="What is 2 + 2?",
            gold_answer="4",
        )

    @pytest.fixture
    def trace(self) -> Trace:
        return Trace(
            problem_id="test_1",
            steps=[
                Step(
                    index=0,
                    operation=StepOperation.LITERAL,
                    before_state="2",
                    after_state="2",
                    output_value=2,
                ),
                Step(
                    index=1,
                    operation=StepOperation.LITERAL,
                    before_state="2",
                    after_state="2",
                    output_value=2,
                ),
                Step(
                    index=2,
                    operation=StepOperation.ADD,
                    before_state="2 + 2",
                    after_state="4",
                    input_refs=[StepRef(step_index=0), StepRef(step_index=1)],
                    output_value=4,
                ),
            ],
        )

    def test_domain_property(self, verifier: ConcreteVerifier):
        """Test domain property."""
        assert verifier.domain == "test_verifier"

    def test_verify_final_correct(self, verifier: ConcreteVerifier, problem: Problem):
        """Test verify_final with correct answer."""
        result = verifier.verify_final(problem, "4")

        assert result.correct is True
        assert result.score == 1.0
        assert result.error_type == ErrorType.NONE
        assert result.expected == "4"
        assert result.actual == "4"

    def test_verify_final_incorrect(self, verifier: ConcreteVerifier, problem: Problem):
        """Test verify_final with incorrect answer."""
        result = verifier.verify_final(problem, "5")

        assert result.correct is False
        assert result.error_type == ErrorType.WRONG_ANSWER
        assert result.expected == "4"
        assert result.actual == "5"

    def test_verify_final_no_gold_answer(self, verifier: ConcreteVerifier):
        """Test verify_final with no gold answer."""
        problem = Problem(
            id="test",
            seed=42,
            domain="test",
            difficulty=DifficultyLevel.EASY,
            prompt="Test",
            gold_answer=None,
        )

        result = verifier.verify_final(problem, "anything")

        assert result.correct is False
        assert result.error_type == ErrorType.INCOMPLETE

    def test_verify_step_correct(
        self,
        verifier: ConcreteVerifier,
        problem: Problem,
        trace: Trace,
    ):
        """Test verify_step with correct value."""
        result = verifier.verify_step(problem, trace, step_index=0, candidate_value=2)

        assert result.correct is True
        assert result.expected == 2
        assert result.actual == 2

    def test_verify_step_incorrect(
        self,
        verifier: ConcreteVerifier,
        problem: Problem,
        trace: Trace,
    ):
        """Test verify_step with incorrect value."""
        result = verifier.verify_step(problem, trace, step_index=0, candidate_value=5)

        assert result.correct is False
        assert result.error_type == ErrorType.WRONG_ANSWER

    def test_verify_step_out_of_range(
        self,
        verifier: ConcreteVerifier,
        problem: Problem,
        trace: Trace,
    ):
        """Test verify_step with out of range index."""
        result = verifier.verify_step(problem, trace, step_index=10, candidate_value=0)

        assert result.correct is False
        assert result.error_type == ErrorType.INVALID_FORMAT
        assert "out of range" in result.error_message.lower()

    def test_verify_step_negative_index(
        self,
        verifier: ConcreteVerifier,
        problem: Problem,
        trace: Trace,
    ):
        """Test verify_step with negative index."""
        result = verifier.verify_step(problem, trace, step_index=-1, candidate_value=0)

        assert result.correct is False
        assert result.error_type == ErrorType.INVALID_FORMAT

    def test_verify_trace_correct(
        self,
        verifier: ConcreteVerifier,
        problem: Problem,
        trace: Trace,
    ):
        """Test verify_trace with correct trace."""
        candidate_trace = Trace(
            problem_id="test_1",
            steps=[
                Step(
                    index=0,
                    operation=StepOperation.LITERAL,
                    before_state="2",
                    after_state="2",
                    output_value=2,
                ),
                Step(
                    index=1,
                    operation=StepOperation.LITERAL,
                    before_state="2",
                    after_state="2",
                    output_value=2,
                ),
                Step(
                    index=2,
                    operation=StepOperation.ADD,
                    before_state="2+2",
                    after_state="4",
                    output_value=4,
                ),
            ],
        )

        result = verifier.verify_trace(problem, trace, candidate_trace)

        assert result.correct is True
        assert result.score == 1.0
        assert result.steps_correct == 3
        assert result.steps_total == 3

    def test_verify_trace_partial_credit(
        self,
        verifier: ConcreteVerifier,
        problem: Problem,
        trace: Trace,
    ):
        """Test verify_trace with partial credit."""
        candidate_trace = Trace(
            problem_id="test_1",
            steps=[
                Step(
                    index=0,
                    operation=StepOperation.LITERAL,
                    before_state="2",
                    after_state="2",
                    output_value=2,
                ),
                Step(
                    index=1,
                    operation=StepOperation.LITERAL,
                    before_state="2",
                    after_state="2",
                    output_value=3,  # Wrong
                ),
                Step(
                    index=2,
                    operation=StepOperation.ADD,
                    before_state="2+2",
                    after_state="5",
                    output_value=5,  # Wrong
                ),
            ],
        )

        result = verifier.verify_trace(problem, trace, candidate_trace)

        assert result.correct is False
        assert result.steps_correct == 1
        assert result.steps_total == 3
        assert result.partial_credit == pytest.approx(1 / 3)
        assert result.first_error_step == 1

    def test_verify_trace_empty(
        self,
        verifier: ConcreteVerifier,
        problem: Problem,
        trace: Trace,
    ):
        """Test verify_trace with empty candidate trace."""
        candidate_trace = Trace(problem_id="test_1", steps=[])

        result = verifier.verify_trace(problem, trace, candidate_trace)

        assert result.correct is False
        assert result.error_type == ErrorType.INCOMPLETE

    def test_verify_trace_shorter_candidate(
        self,
        verifier: ConcreteVerifier,
        problem: Problem,
        trace: Trace,
    ):
        """Test verify_trace with shorter candidate trace."""
        candidate_trace = Trace(
            problem_id="test_1",
            steps=[
                Step(
                    index=0,
                    operation=StepOperation.LITERAL,
                    before_state="2",
                    after_state="2",
                    output_value=2,
                ),
            ],
        )

        result = verifier.verify_trace(problem, trace, candidate_trace)

        # Should only check available steps
        assert result.steps_correct == 1
        assert result.steps_total == 3

    def test_grade_tool_usage_default(
        self,
        verifier: ConcreteVerifier,
        problem: Problem,
    ):
        """Test default grade_tool_usage."""
        tool_calls = [
            {"name": "calculator", "args": {"expr": "2+2"}},
            {"name": "verify", "args": {"value": 4}},
        ]

        grades = verifier.grade_tool_usage(problem, tool_calls)

        assert len(grades) == 2
        assert all(g.valid for g in grades)
        assert all(g.necessary for g in grades)
        assert all(g.efficient for g in grades)
        assert grades[0].tool_name == "calculator"
        assert grades[1].tool_name == "verify"

    def test_grade_tool_usage_empty(
        self,
        verifier: ConcreteVerifier,
        problem: Problem,
    ):
        """Test grade_tool_usage with no tool calls."""
        grades = verifier.grade_tool_usage(problem, [])

        assert grades == []

    def test_grade_tool_usage_missing_name(
        self,
        verifier: ConcreteVerifier,
        problem: Problem,
    ):
        """Test grade_tool_usage with missing name."""
        tool_calls = [{"args": {"x": 1}}]

        grades = verifier.grade_tool_usage(problem, tool_calls)

        assert len(grades) == 1
        assert grades[0].tool_name == "unknown"

    def test_classify_error_default(
        self,
        verifier: ConcreteVerifier,
        problem: Problem,
    ):
        """Test default classify_error."""
        error_type = verifier.classify_error(problem, expected=4, actual=5)
        assert error_type == ErrorType.WRONG_ANSWER

    def test_classify_error_none(
        self,
        verifier: ConcreteVerifier,
        problem: Problem,
    ):
        """Test classify_error with None actual."""
        error_type = verifier.classify_error(problem, expected=4, actual=None)
        assert error_type == ErrorType.INCOMPLETE

    def test_classify_error_custom(
        self,
        numeric_verifier: NumericVerifier,
        problem: Problem,
    ):
        """Test custom classify_error."""
        # Sign error
        error_type = numeric_verifier.classify_error(problem, expected=4, actual=-4)
        assert error_type == ErrorType.SIGN_ERROR

        # Rounding error
        error_type = numeric_verifier.classify_error(problem, expected=100, actual=100.5)
        assert error_type == ErrorType.ROUNDING_ERROR

    def test_parse_answer_default(
        self,
        verifier: ConcreteVerifier,
        problem: Problem,
    ):
        """Test default parse_answer."""
        parsed = verifier.parse_answer(problem, "  42  ")
        assert parsed == "42"

    def test_format_answer_default(
        self,
        verifier: ConcreteVerifier,
        problem: Problem,
    ):
        """Test default format_answer."""
        formatted = verifier.format_answer(problem, 42)
        assert formatted == "42"


class TestNumericVerifier:
    """Tests for NumericVerifier with tolerance."""

    @pytest.fixture
    def verifier(self) -> NumericVerifier:
        return NumericVerifier()

    @pytest.fixture
    def problem(self) -> Problem:
        return Problem(
            id="test_1",
            seed=42,
            domain="numeric",
            difficulty=DifficultyLevel.MEDIUM,
            prompt="Calculate pi",
            gold_answer="3.14159",
            answer_tolerance=0.001,
        )

    def test_verify_final_within_tolerance(self, verifier: NumericVerifier, problem: Problem):
        """Test numeric verification within tolerance."""
        result = verifier.verify_final(problem, "3.1416")

        assert result.correct is True

    def test_verify_final_outside_tolerance(self, verifier: NumericVerifier, problem: Problem):
        """Test numeric verification outside tolerance."""
        result = verifier.verify_final(problem, "3.2")

        assert result.correct is False

    def test_verify_final_invalid_format(self, verifier: NumericVerifier, problem: Problem):
        """Test numeric verification with invalid format."""
        result = verifier.verify_final(problem, "not a number")

        assert result.correct is False
        assert result.error_type == ErrorType.INVALID_FORMAT
